/*----------------------------------------------------------------------
  File    : gramod4.c
  Contents: graphical model management (based on directed graphs)
            network induction based on conditional independence tests
  Author  : Christian Borgelt
  History : 1995.11.03 file created as ines.c
            2002.03.03 network induction functions moved to gramod2.c
            2002.07.04 local structure learning (re)added
            2006.03.04 cond. independence tests construction added
            2006.03.05 function condep() completed (later: reduce())
            2006.03.06 some omissions in function reduce() fixed
            2006.03.08 started to debug function reduce()
            2006.03.09 memory leaks in reduce() fixed
            2006.03.10 tests in needed() and reduce() extended
            2006.03.11 bug in dired() fixed (common neighbors)
            2006.03.12 most fprintf replaced by fputs and fputc
            2006.12.20 induction functions simplified
            2007.01.28 induction of undirected graphs added (_undired)
            2007.01.29 bug in function dired() fixed (number of conds.)
            2007.06.11 function bestsep() simplified (common neighbors)
            2013.09.04 adapted to definitions ATTID, VALID, TPLID etc.
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include "gmutil.h"
#include "arrays.h"
#ifndef GR_EXTFN
#define GR_EXTFN
#endif
#include "graph.h"
#include "citrie.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Conditional Independence Tests Construction
----------------------------------------------------------------------*/

static void logout (GMID *gmid, GMEDGE *edge,
                    ATTID *cons, ATTID cnt, double eval)
{                               /* --- write evaluation to log file */
  ATTID i;                      /* loop variable */

  assert(gmid                   /* check the function arguments */
  &&     edge && (cons || (cnt <= 0)));
  fputs(att_name(tab_col(gmid->table, edge->x)), gmid->log);
  fputc(' ',   gmid->log);      /* print first  attribute */
  fputs(att_name(tab_col(gmid->table, edge->y)), gmid->log);
  fputs(" | ", gmid->log);      /* print second attribute */
  for (i = 0; i < cnt; i++) {   /* traverse the conditions */
    fputs(att_name(tab_col(gmid->table, cons[i])), gmid->log);
    fputc(' ', gmid->log);      /* print list of conditions */
  }                             /* finally print the evaluation */
  fprintf(gmid->log, " : %g\n", eval);
}  /* logout() */

/*--------------------------------------------------------------------*/

static PTREE* ptree (GMID *gmid, GMEDGE *edge, ATTID *cons, ATTID cnt)
{                               /* --- tree for cond. indep. test */
  ATTID i;                      /* loop variable */
  PTREE *pt;                    /* created prob./poss. tree */

  assert(gmid                   /* check the function arguments */
  &&     edge && (cons || (cnt <= 0)));
  pt = pt_create(tab_col(gmid->table, edge->x), gmid->gm->type);
  if (!pt) return NULL;         /* create prob./poss. tree */
  if (pt_conadd(pt, 0, tab_col(gmid->table, edge->y)) < 0) {
    pt_delete(pt, 0); return NULL; }    /* add other attribute */
  for (i = 0; i < cnt; i++)     /* and then add all conditions */
    if (pt_conadd(pt, i, tab_col(gmid->table, cons[i])) < 0) {
      pt_delete(pt, 0); return NULL; }
  return pt;                    /* return the created tree */
}  /* ptree() */

/*--------------------------------------------------------------------*/

static int citest (GMID *gmid, GMEDGE *edge, ATTID *cons, ATTID cnt)
{                               /* --- execute a cond. indep. test */
  PTREE  *pt;                   /* prob./poss. tree for the test */
  double eval;                  /* evaluation of cond. dependence */
  CITRIE *cit;                  /* to access the cond. indep. trie */

  assert(gmid                   /* check the function arguments */
  &&     edge && (cons || (cnt <= 0)));
  cit  = (CITRIE*)gmid->citrie; /* get the cond. independence trie */
  eval = cit_get(cit, edge->x, edge->y, cons, cnt);
  if (isnan(eval)) {            /* if the evaluation is not known yet */
    pt = ptree(gmid, edge, cons, cnt);
    if (!pt) return -1;         /* create initial tree for c.i. test */
    if (gmu_ptaggr(&pt, 1, gmid->table) != 0) {
      pt_delete(pt, 0); return -1; }  /* aggregate tuples in tree */
    eval = pt_depend(pt, gmid->measure, gmid->params);
    pt_delete(pt, 0);           /* evaluate the prob./poss. tree, */
    if (isnan(eval)             /* then delete the evaluated tree */
    || (cit_add(cit, edge->x, edge->y, cons, cnt, eval) != 0))
      return -1;                /* store the evaluation in the trie */
    if (gmid->log) logout(gmid, edge, cons, cnt, eval);
  }                             /* write evaluation to the log file */
  edge->eval = eval;            /* store the evaluation */
  return 0;                     /* return 'ok' */
}  /* citest() */

/*--------------------------------------------------------------------*/

static ATTID reduce (GMID *gmid, GMEDGE *edge,
                     ATTID *cons, ATTID cnt, ATTID rem, int full)
{                               /* --- reduce set of conditions */
  ATTID  i, n;                  /* loop variables, tree counter */
  PTREE  *pt, **trees;          /* to traverse the prob. trees */
  ATTID  curr, best;            /* current and best condition */
  double eval, *evec;           /* evaluation of cond. dependence */
  CITRIE *cit;                  /* to access the cond. indep. trie */

  assert(gmid                   /* check the function arguments */
  &&     edge && (cons || (cnt <= 0)) && (cnt >= rem) && (rem >= 0));

  /* --- test full set of conditions --- */
  if (citest(gmid, edge, cons, cnt) < 0)
    return -1;                  /* determine the initial evaluation */
  eval = edge->eval;            /* (with all given conditions) */
  if ((!full && (eval <= gmid->thresh))
  ||  (rem <= 0))               /* if conditional independence found */
    return cnt;                 /* or no condition to remove, abort */

  /* --- greedy removal of conditions --- */
  trees = (PTREE**)calloc((size_t)rem, sizeof(PTREE*));
  if (!trees) return -1;        /* allocate some working memory */
  evec = (double*)malloc((size_t)rem *sizeof(double));
  if (!evec) { free(trees); return -1; }
  cit = (CITRIE*)gmid->citrie;  /* get the cond. independence trie */
  for ( ; rem > 0; --rem) {     /* while conditions can be removed */
    n = 0; --cnt;               /* reduce the number of conditions */
    for (i = rem; --i >= 0; ) { /* traverse the removable conditions */
      curr    = cons[i];        /* replace the current condition */
      cons[i] = cons[cnt];      /* with the last one in the set */
      evec[i] = cit_get(cit, edge->x, edge->y, cons, cnt);
      if (isnan(evec[i])) {     /* if the evaluation is unknown */
        pt = ptree(gmid, edge, cons, cnt);
        if (!pt) return gmu_ptclean(trees, rem, evec);
        trees[n++] = pt;        /* create a corresp. prob./poss. tree */
      }                         /* and add it to the array of trees */
      cons[i] = curr;           /* restore the removed condition */
    }                           /* (keep condition set consistent) */
    if ((n > 0)                 /* aggregate the tuples for all trees */
    &&  (gmu_ptaggr(trees, n, gmid->table) != 0))
      return gmu_ptclean(trees, rem, evec);
    best = -1; n = 0;           /* init. index of best condition */
    for (i = rem; --i >= 0; ) { /* traverse the prob./poss. trees */
      curr    = cons[i];        /* replace the removed condition */
      cons[i] = cons[cnt];      /* with the last one in the set */
      if (isnan(evec[i])) {     /* if the evaluation is unknown */
        evec[i] = pt_depend(trees[n], gmid->measure, gmid->params);
        pt_delete(trees[n], 0); /* evaluate the prob./poss. tree and */
        trees[n++] = NULL;      /* store the evaluation in the trie */
        if (isnan(evec[i])
        || (cit_add(cit, edge->x, edge->y, cons, cnt, eval) != 0))
          return gmu_ptclean(trees, rem, evec);
        if (gmid->log) logout(gmid, edge, cons, cnt, evec[i]);
      }                         /* write evaluation to the log file */
      if (evec[i] < eval)       /* if a better evaluation was found, */
        eval = evec[best = i];  /* store it and the condition index */
      cons[i] = curr;           /* restore the removed condition */
    }                           /* (keep condition set consistent) */
    if (best < 0) break;        /* if no removal improves, abort */
    for (i = best; ++i <= cnt;) /* remove the condition that */
      cons[i-1] = cons[i];      /* produced the best evaluation */
    if (!full && (eval <= gmid->thresh))
      break;                    /* if conditional independence found, */
  }                             /* abort the condition removal */
  edge->eval = eval;            /* store the best evaluation */
  free(trees); free(evec);      /* delete the working memory */
  return cnt+1;                 /* return the number of consitions */
}  /* reduce() */

/*--------------------------------------------------------------------*/

static ATTID bestsep (GMID *gmid, GRAPH *graph, GMEDGE *edge)
{                               /* --- find best separating conds. */
  ATTID i, n1, n2;              /* loop variable, numbers of nodes */
  ATTID *adjs1, *adjs2;         /* vectors of adjacency neighbors */

  assert(gmid && graph && edge);/* check the function arguments */
  adjs1 = gmid->adjs;           /* get the two buffers */
  adjs2 = adjs1 +gmid->attcnt;  /* for the node identifiers */
  n1 = gr_transto(graph, edge->x, edge->y, 0, adjs1);
  n2 = gr_transto(graph, edge->x, edge->y, 0, adjs2);
  if (n2 < n1) {                /* get the transitive neighbors */
    for (i = n1 = n2; --i >= 0; ) adjs1[i] = adjs2[i]; }
  return reduce(gmid, edge, adjs1, n1, n1, -1);
}  /* bestsep() */              /* reduce the set of conditions */

/*----------------------------------------------------------------------
Note that in the function bestsep() the function reduce() is called
with the parameter full = -1, thus causing the set of conditions to be
shrunk as long as the evaluation gets better by removing a condition.
In contrast to this function needed() calls the function reduce() with
full = 0, thus causing it to return if a conditional independence was
found, even if a further removal of conditions could improve the score.
----------------------------------------------------------------------*/

static int needed (GMID *gmid, GRAPH *graph, GMEDGE *edge, int mode)
{                               /* --- check whether edge is needed */
  ATTID n1, n2, k;              /* numbers of nodes, index/buffer */
  ATTID *adjs1, *adjs2;         /* vectors of adjacency neighbors */

  assert(gmid && graph && edge);/* check the function arguments */
  adjs1 = gmid->adjs;           /* get the two buffers */
  adjs2 = adjs1 +gmid->attcnt;  /* for the node identifiers */

  /* --- undirected and heuristic test --- */
  if (mode <= 0) {              /* if to do a heuristic test */
    n1 = gr_adjsto(graph, edge->x, edge->y, 0, adjs1);
    n2 = gr_adjsto(graph, edge->y, edge->x, 0, adjs2);
    if (n2 < n1) {              /* get sets of adjacency neighbors */
      adjs1 = adjs2; adjs2 = gmid->adjs;
      k = n1; n1 = n2; n2 = k;  /* make the smaller set of nodes */
    }                           /* the one to be tested first */
    k = (mode < 0) ? citest(gmid, edge, adjs1, n1)
                   : reduce(gmid, edge, adjs1, n1, n1, 0);
    if (k < 0) return -1;       /* try first  set of neighbors */
    if (edge->eval <= gmid->thresh)
      return  0;                /* check for conditional independence */
    k = (mode < 0) ? citest(gmid, edge, adjs2, n2)
                   : reduce(gmid, edge, adjs2, n2, n2, 0);
    if (k < 0) return -1;       /* try second set of neighbors */
    return (edge->eval <= gmid->thresh) ? 0 : 1;
  }                             /* check for conditional independence */

  /* --- strict test --- */
  n1 = gr_transto(graph, edge->x, edge->y, 0, adjs1);
  n2 = gr_transto(graph, edge->x, edge->y, 0, adjs2);
  if (n2 < n1) {                /* get the transitive neighbors */
    adjs1 = adjs2; adjs2 = gmid->adjs; n1 = n2; }
  if (reduce(gmid, edge, adjs1, n1, n1, 0) < 0)
    return -1;                  /* reduce set of conditions */
  return (edge->eval <= gmid->thresh) ? 0 : 1;
}  /* needed() */               /* check for conditional independence */

/*----------------------------------------------------------------------
The parameter 'mode' of the above function has the following meaning:
mode = -1: do not reduce the set of conditions (for undirected tests)
mode =  0: heuristic test, reduce the set of conditions
mode =  1: strict    test, reduce the set of conditions
----------------------------------------------------------------------*/

static GRAPH* drafting (GMID *gmid)
{                               /* --- draft graph as spanning tree */
  ATTID  i, k;                  /* loop variables, buffers */
  GMEDGE **p;                   /* to traverse the edges */
  GRAPH  *graph;                /* created graph structure */

  assert(gmid);                 /* check the function argument */
  if ((gmu_edges (gmid) < 0)    /* collect the selectable edges */
  ||  (gmu_eval  (gmid, -1, -1) < 0)
  ||  (gmu_filter(gmid) < 0)    /* evaluate and filter the edges */
  ||  (gmu_span  (gmid) < 0))   /* and construct a spanning tree */
    return NULL;                /* from the remaining edges */
  p = gmid->edges;              /* traverse the initial edges */
  for (i = gmid->inicnt; --i >= 0; p++)
    if (cit_add(gmid->citrie, (*p)->x,(*p)->y,NULL,0,(*p)->eval) != 0)
      return NULL;              /* store unconditional dependences */
  graph = gr_createx(tab_attset(gmid->table));
  if (!graph) return NULL;      /* create a graph to work on */
  p = gmid->edges;              /* traverse the tree edges */
  for (i = gmid->selcnt; --i >= 0; p++) {
    if (((*p)->mark < 2)        /* process only the added edges */
    &&  (gr_addundir(graph, (*p)->x, (*p)->y) != 0)) {
      gr_delete(graph); return NULL; }
  }                             /* transfer spanning tree to graph */
  for (i = gmid->attcnt; --i >= 0; ) {
    for (k = gm_concnt(gmid->gm, i); --k >= 0; ) {
      if (gr_addedge(graph, gm_conid(gmid->gm, i, k), i) != 0) {
        gr_delete(graph); return NULL; }
    }                           /* transfer the existing conditions */
  }                             /* as edges to the created graph */
  return graph;                 /* return the created graph */
}  /* drafting() */

/*--------------------------------------------------------------------*/

static int clqs2gm (GMID *gmid, GRAPH *graph)
{                               /* --- build graphical model */
  ATTID i, k, n, c;             /* loop variables, buffers */
  ATTID *clique, *marks;        /* to traverse the cliques */

  gm_conrem(gmid->gm, -1, 0);   /* remove all conditions */
  clique = gmid->adjs; n = 0;   /* initialize the clique buffer */
  marks  = gmid->adjs +(i = gmid->attcnt);
  while (--i >= 0) marks[i] = 0;/* clear all node markers */
  while (1) {                   /* maximal clique retrieval loop */
    n = gr_maxclq(graph, clique, n);
    if (n <= 0) break;          /* get the next maximal clique */
    if (n <= 1) continue;       /* ignore isolated nodes */
    if (gmid->log) {            /* if a log file is given */
      for (i = 0; i < n; i++) { /* print the maximal clique */
        fputs(att_name(tab_col(gmid->table, clique[i])), gmid->log);
        fputc(' ', gmid->log);  /* print the node names */
      }                         /* separated by spaces */
    }
    for (i = 0; i < n; i++)     /* try to find an unused node */
      if (!marks[clique[i]]) break;
    if (i < 0) {                /* if there is no unused node */
      if (gmid->log) fputs("(skipped)\n", gmid->log);
      continue;                 /* skip the clique */
    }
    marks[c = clique[i]] = 1;   /* mark the selected node as used */
    if (gmid->log) {            /* if a log file is given, */
      fputs("-> ", gmid->log);  /* print a selection indication */
      fputs(att_name(tab_col(gmid->table, clique[i])), gmid->log);
      fputc('\n', gmid->log);   /* print the selected attribute */
    }
    for (k = n; --k >= 0; ) {   /* traverse the nodes in the clique */
      if (k == i) continue;     /* except the chosen node */
      if (gm_conadd(gmid->gm, c, clique[k]) < 0) return -1;
    }                           /* add all other nodes in the clique */
  }                             /* as conditions to the model */
  return 0;                     /* return 'ok' */
}  /* clqs2gm() */

/*--------------------------------------------------------------------*/

static int dired (GMID *gmid)
{                               /* --- directed graph induction */
  ATTID       i, k, n, a, b, c; /* loop variables, buffers */
  GMEDGE      **p, *e;          /* to traverse the edges */
  const ATTID *cons;            /* identifiers of condition nodes */
  ATTID       *common;          /* list of common neighbors */
  GRAPH       *graph;           /* created graph structure */

  assert(gmid);                 /* check the function argument */

  /* --- drafting --- */
  fputs(" drafting\b\b\b\b\b\b\b\b\b", stderr);
  if (gmid->log) fputs("drafting:\n", gmid->log);
  graph = drafting(gmid);       /* draft the initial graph */
  if (!graph) return -1;        /* as an optimum spanning tree */
  if (gmid->log) {              /* describe the graph in the log */
    fputc('\n', gmid->log); gr_desc(graph, gmid->log, GR_NAMES, 0); }

  /* --- thickening --- */
  fputs(" thickening\b\b\b\b\b\b\b\b\b\b\b", stderr);
  if (gmid->log) fputs("\nthickening:\n", gmid->log);
  gmid->adjs = (ATTID*)malloc(3 *(size_t)gmid->attcnt *sizeof(ATTID));
  if (!gmid->adjs) { gr_delete(graph); return -1; }
  p = gmid->edges +gmid->selcnt;/* get the edges that were not */
  n = gmid->potcnt-gmid->selcnt;/* selected and their number */
  ptr_qsort(p, (size_t)n, 1, gmu_cmp, NULL);
  for ( ; --n >= 0; p++) {      /* resort these remaining edges */
    if ((*p)->mark != 1) continue;
    k = needed(gmid, graph, *p, 0);
    if (k == 0) (*p)->mark = 0; /* unmark all unnecessary edges */
    if (k >  0) k = gr_addundir(graph, (*p)->x, (*p)->y);
    if (k <  0) { gr_delete(graph); return -1; }
  }                             /* add edges that appear to be needed */
  if (gmid->log) {              /* describe the graph in the log */
    fputc('\n', gmid->log); gr_desc(graph, gmid->log, GR_NAMES, 0); }

  /* --- thinning heuristic --- */
  fputs(" thinning heuristic", stderr);
  fputs("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b", stderr);
  if (gmid->log) fputs("\nthinning 1:\n", gmid->log);
  p = gmid->edges;              /* traverse all potential edges */
  for (p += i = gmid->potcnt; --i >= 0; ) {
    if ((*--p)->mark != 1) continue;
    k = needed(gmid, graph, *p, 0);
    if (k <  0) { gr_delete(graph); return -1; }
    if (k == 0) { gr_remundir(graph, (*p)->x, (*p)->y);
                  (*p)->mark = 0; }
  }                             /* remove unnecessary edges */
  if (gmid->log) {              /* describe the graph in the log */
    fputc('\n', gmid->log); gr_desc(graph, gmid->log, GR_NAMES, 0); }

  /* --- thinning strict --- */
  if (gmid->search == GM_CITEST) {
    fputs(" thinning strict   ", stderr);
    fputs("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b", stderr);
    if (gmid->log) fputs("\nthinning 2:\n", gmid->log);
    p = gmid->edges;            /* traverse all potential edges */
    for (p += i = gmid->potcnt; --i >= 0; ) {
      if ((*--p)->mark != 1) continue;
      k = needed(gmid, graph, *p, 1);
      if (k <  0) { gr_delete(graph); return -1; }
      if (k == 0) { gr_remundir(graph, (*p)->x, (*p)->y);
                    (*p)->mark = 0; }
    }                           /* remove unnecessary edges */
    if (gmid->log) {            /* describe the graph in the log */
      fputc('\n', gmid->log); gr_desc(graph, gmid->log, GR_NAMES, 0); }
  }                             /* (strict thinning may be omitted) */

  /* --- orient edges (v-structures) --- */
  fputs(" orienting      ", stderr);
  fputs("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b", stderr);
  if (gmid->log) fputs("\norienting 1 (v-structures):\n", gmid->log);
  common = gmid->adjs +2 *gmid->attcnt;
  for (i = gmid->attcnt; --i > 0; ) {
    for (e = gmid->etab[i] +(k = i); --k >= 0; ) {
      if ((--e)->mark != 0)     /* traverse all node pairs, but */
        continue;               /* skip pairs of adjacent nodes */
      c = bestsep(gmid, graph, e);
      if (c < 0) {              /* get best separating conditions */
        gr_delete(graph); return -1; }
      n = gr_common(graph, i, k, GR_NOPARENT, common);
      while (--n >= 0) {        /* get the common neighbors */
        a = common[n];          /* traverse them */
        for (b = c; --b >= 0; ) /* traverse the best conditions */
          if (gmid->adjs[b] == a) break;
        if (b >= 0)             /* if the common neighbor is in the */
          continue;             /* condition set -> no v-structure */
        if ((gr_dirpath(graph, a, i) < 0)
        ||  (gr_dirpath(graph, a, k) < 0)) {
          if (gmid->log) {      /* do not introduce a directed cycle */
            fputs("skipped v-structure: ",           gmid->log);
            fputs(att_name(tab_col(gmid->table, i)), gmid->log);
            fputs(" -> ",                            gmid->log);
            fputs(att_name(tab_col(gmid->table, a)), gmid->log);
            fputs(" <- ",                            gmid->log);
            fputs(att_name(tab_col(gmid->table, k)), gmid->log);
            fputc('\n',                              gmid->log);
          } continue;           /* report skipped v-structures */
        }                       /* in the log file (if there is one) */
        gr_remdir(graph, a, i); /* otherwise remove edge directions */
        gr_remdir(graph, a, k); /* away from the common neighbor, */
      }                         /* so that both edges point from */
    }                           /* the nodes to the common neighbor */
  }                             /* (create a v-structure/collider) */
  if (gmid->log) {              /* describe the graph in the log */
    fputc('\n', gmid->log); gr_desc(graph, gmid->log, GR_NAMES, 0); }

  /* --- orient edges (remaining edges) --- */
  if (gmid->log) {              /* if a log file is given */
    fputs("\norienting 2 (remaining edges):\n", gmid->log);
    for (i = gmid->attcnt; --i >= 0; )
      gmid->adjs[i] = gr_parcnt(graph, i);
  }                             /* not the number of parents */
  gr_direct(graph, 1);          /* fully direct the graph */
  if (gmid->log) {              /* if a log file is given */
    a = 0;                      /* flag for "no edge printed yet" */
    for (i = gmid->attcnt; --i > 0; ) {
      n = gr_parcnt(graph, i);  /* traverse all nodes */
      if ((n <= 1) || (gmid->adjs[i] >= n))
        continue;               /* check possible new v-structures */
      cons = gr_parents(graph, i);
      for (k = gmid->adjs[i]; k < n; k++) {
        c = cons[k];            /* traverse the new parents */
        for (b = n; --b >= 0; ) /* traverse all other parents */
          if ((b != k) && (gr_edge(graph, c, cons[b]) == 0))
            break;              /* check for a connecting edge */
        if (b < 0) continue;    /* (harmless new v-structures) */
        if (a >= 0) { a = -1;   /* print header for log-file */
          fputs("edges creating new v-structures:\n", gmid->log); }
        fputs(att_name(tab_col(gmid->table, c)), gmid->log);
        fputs(" -> ",                            gmid->log);
        fputs(att_name(tab_col(gmid->table, i)), gmid->log);
        fputc('\n',                              gmid->log);
      }                         /* print all edges that */
    }                           /* cause new v-structures */
    fputc('\n', gmid->log); gr_desc(graph, gmid->log, GR_NAMES, 0);
  }                             /* describe the final graph */

  /* --- moralize the graph --- */
  if (gmid->gm->type == GM_POSS) {
    fputs(" moralizing \b\b\b\b\b\b\b\b\b\b\b\b", stderr);
    if (gmid->log) fputs("\nmoralizing:\n", gmid->log);
    gr_moralize(graph);         /* moralize the final graph */
    gr_undirect(graph);         /* and remove all directions */
    if (gmid->log) {            /* if a log file is given */
      fputc('\n', gmid->log); gr_desc(graph, gmid->log, GR_NAMES, 0); }
  }                             /* describe the moralized graph */

  /* --- build graphical model --- */
  fputs(" finishing \b\b\b\b\b\b\b\b\b\b\b", stderr);
  if (gmid->log) fputs("\nfinishing:\n", gmid->log);
  gm_conrem(gmid->gm, -1, 0);   /* remove all conditions */
  if (gmid->gm->type == GM_POSS)/* if possibilistic graphical model, */
    n = clqs2gm(gmid, graph);   /* build from the maximal cliques */
  else {                        /* if to build direcly */
    for (n = 0, i = gmid->attcnt; (--i >= 0) && (n == 0); ) {
      cons = gr_parents(graph,i); /* traverse the nodes */
      for (k = gr_parcnt(graph, i); --k >= 0; )
        if (gm_conadd(gmid->gm, i, cons[k]) < 0) { n = -1; break; }
    }                           /* traverse the parent nodes and */
  }                             /* add all parents as conditions */
  gr_delete(graph);             /* delete the graph structure */
  if (n != 0) return -1;        /* check for an error */
  if (gmid->log) {              /* if a log file is given */
    fputs("\ntest statistics:\n",    gmid->log);
    cit_stats((CITRIE*)gmid->citrie, gmid->log);
  }                             /* print cond. indep. test statistics */
  return 0;                     /* return 'ok' */
}  /* dired() */

/*--------------------------------------------------------------------*/

static int undired (GMID *gmid)
{                               /* --- undirected graph induction */
  ATTID  i, k, n;               /* loop variables, buffers */
  GMEDGE **p, *e;               /* to traverse the edges */
  GRAPH  *graph;                /* created graph structure */

  assert(gmid);                 /* check the function argument */

  /* --- drafting --- */
  fputs(" drafting\b\b\b\b\b\b\b\b\b", stderr);
  if (gmid->log) fputs("drafting:\n", gmid->log);
  graph = drafting(gmid);       /* draft the initial graph */
  if (!graph) return -1;        /* as an optimum spanning tree */
  gr_undirect(graph);           /* undirect all directed edges */
  if (gmid->log) {              /* describe the graph in the log */
    fputc('\n', gmid->log); gr_desc(graph, gmid->log, GR_NAMES, 0); }

  /* --- thickening --- */
  fputs(" thickening\b\b\b\b\b\b\b\b\b\b\b", stderr);
  if (gmid->log) fputs("\nthickening:\n", gmid->log);
  gmid->adjs = (ATTID*)malloc(2 *(size_t)gmid->attcnt *sizeof(ATTID));
  if (!gmid->adjs) { gr_delete(graph); return -1; }
  p = gmid->edges +gmid->selcnt;/* get the edges that were not */
  n = gmid->potcnt-gmid->selcnt;/* selected and their number */
  ptr_qsort(p, (size_t)n, 1, gmu_cmp, NULL);
  for ( ; --n >= 0; p++) {      /* resort these remaining edges */
    if ((*p)->mark != 1) continue;
    k = needed(gmid, graph, *p, -1);
    if (k == 0) (*p)->mark = 0; /* unmark all unnecessary edges */
    if (k >  0) k = gr_addundir(graph, (*p)->x, (*p)->y);
    if (k <  0) { gr_delete(graph); return -1; }
  }                             /* add edges that appear to be needed */
  if (gmid->log) {              /* describe the graph in the log */
    fputc('\n', gmid->log); gr_desc(graph, gmid->log, GR_NAMES, 0); }

  /* --- thinning --- */
  if (gmid->search == GM_CIUNDIR) {
    fputs(" thinning  \b\b\b\b\b\b\b\b\b\b\b", stderr);
    if (gmid->log) fputs("\nthinning:\n", gmid->log);
    p = gmid->edges;            /* traverse all potential edges */
    for (p += i = gmid->potcnt; --i >= 0; ) {
      if ((*--p)->mark != 1) continue;
      k = needed(gmid, graph, *p, -1);
      if (k <  0) { gr_delete(graph); return -1; }
      if (k == 0) { gr_remundir(graph, (*p)->x, (*p)->y);
                    (*p)->mark = 0; }
    }                           /* remove unnecessary edges */
    if (gmid->log) {            /* describe the graph in the log */
      fputc('\n', gmid->log); gr_desc(graph, gmid->log, GR_NAMES, 0); }
  }                             /* (this thinning may be skipped) */

  /* --- moralizing --- */
  fputs(" moralizing\b\b\b\b\b\b\b\b\b\b\b", stderr);
  if (gmid->log) fputs("\nmoralizing:\n", gmid->log);
  for (n = gmid->attcnt; --n > 0; ) {
    for (e = gmid->etab[n] +(i = n); --i >= 0; ) {
      if ((--e)->mark != 0)     /* traverse all node pairs, but */
        continue;               /* skip pairs of adjacent nodes */
      assert((e->x == n) && (e->y == i));
      if (gr_common(graph, n, i, 0, NULL) <= 0)
        continue;               /* skip nodes w/o common neighbors */
      k = needed(gmid, graph, e, -1);
      if (k < 0) { gr_delete(graph); return -1; }
      if (k > 0) e->mark = -1;  /* mark all moralizing edges */
    }                           /* that need to be added */
  }                             /* (but add them only later) */
  for (n = gmid->attcnt; --n > 0; ) {
    for (e = gmid->etab[n] +(i = n); --i >= 0; ) {
      if ((--e)->mark >= 0) continue;
      e->mark = 1;              /* traverse the marked edges */
      assert((e->x == n) && (e->y == i));
      if (gr_addundir(graph, n, i) < 0) {
        gr_delete(graph); return -1; }
    }                           /* actually add all marked edges */
  }                             /* to the graph */
  if (gmid->log) {              /* describe the graph in the log */
    fputc('\n', gmid->log); gr_desc(graph, gmid->log, GR_NAMES, 0); }

  /* --- thinning --- */
  fputs(" thinning  \b\b\b\b\b\b\b\b\b\b\b", stderr);
  if (gmid->log) fputs("\nthinning:\n", gmid->log);
  p = gmid->edges;              /* traverse all potential edges */
  for (p += i = gmid->potcnt; --i >= 0; ) {
    if ((*--p)->mark != 1) continue;
    k = needed(gmid, graph, *p, -1);
    if (k <  0) { gr_delete(graph); return -1; }
    if (k == 0) { gr_remundir(graph, (*p)->x, (*p)->y);
                  (*p)->mark = 0; }
  }                             /* remove unnecessary edges */
  if (gmid->log) {              /* describe the graph in the log */
    fputc('\n', gmid->log); gr_desc(graph, gmid->log, GR_NAMES, 0); }

  /* --- build graphical model --- */
  fputs(" finishing \b\b\b\b\b\b\b\b\b\b\b", stderr);
  if (gmid->log) fputs("\nfinishing:\n", gmid->log);
  k = clqs2gm(gmid, graph);     /* build the graphical model */
  gr_delete(graph);             /* delete the graph structure */
  if (k != 0) return -1;        /* check for an error */
  if (gmid->log) {              /* if a log file is given */
    fputs("\ntest statistics:\n",    gmid->log);
    cit_stats((CITRIE*)gmid->citrie, gmid->log);
  }                             /* print cond. indep. test statistics */
  return 0;                     /* return 'ok' */
}  /* undired() */

/*--------------------------------------------------------------------*/

double gm_citest (GRAMOD *gm, TABLE *table, int search, int measure,
                  double *params, double thresh, int lwise,
                  double mrgimp, ATTID maxcon, ATTID maxtest, FILE *log)
{                               /* --- induce by selecting conditions */
  double tplcnt;                /* number of tuples (weight) */
  GMID   gmid;                  /* structure of induction variables */
  int    err;                   /* whether an error occurred */

  assert(gm && table);          /* check the function arguments */
  if (gm->attcnt <= 0)          /* if there are no attributes, */
    return 0.0;                 /* simply abort the function */
  tplcnt = tab_getwgt(table, 0, TPLID_MAX);
  if ((gm->type & GM_POSS)      /* for poss. graphical models */
  &&  (tab_opc(table, TAB_COND) != 0))
    return NAN;                 /* compute one point coverages */
  gm_setnorm(gm, tplcnt);       /* set number of tuples for norm. */
  assert((search >= GM_CITEST) && (search <= GM_CIUDNIT));
  gmid.gm        = gm;          /* initialize the */
  gmid.table     = table;       /* induction variables */
  gmid.search    = search;
  gmid.measure   = measure;
  gmid.params[0] = params[0];
  gmid.params[1] = params[1];
  gmid.thresh    = (thresh  <= 0) ? 16*DBL_EPSILON : thresh;
  gmid.mrgimp    =  0;
  gmid.lwise     = -1;
  gmid.maxcon    = (maxcon  >= gm->attcnt) ? gm->attcnt-2 : maxcon;
  gmid.ptmax     = (maxtest <= 0) ? 1 : maxtest;
  gmid.log       = log;
  if (gmu_init(&gmid, gm->attcnt, -1) < 0)
    return NAN;                 /* init. edge table (half, no diag.) */
  gmid.citrie = cit_create(gm->attcnt);
  if (!gmid.citrie) return NAN; /* create a trie for cond. indeps. */
  err = (search >= GM_CIUNDIR)  /* induce the graphical model */
      ? undired(&gmid) : dired(&gmid);
  if (gmid.citrie) cit_delete((CITRIE*)gmid.citrie);
  gmu_cleanup(&gmid);           /* clean up the induction variables */
  if (err) return NAN;          /* and check for an error */
  gm_clear(gm, 1);              /* clear the graphical model */
  if ((gm_estim(gm, table, 0) != 0)
  ||  isnan(gm_eval(gm, measure, params)))
    return NAN;                 /* rebuild the graphical model */
  if ((lwise >= 0) && isnan(gm_local(gm, lwise, mrgimp)))
    return NAN;                 /* redo local structure learning */
  return gm->eval;              /* return the evaluation result */
}  /* gm_citest() */
