/*----------------------------------------------------------------------
  File    : hyper.c
  Contents: hypertree management for simulated annealing
  Author  : Christian Borgelt
  History : 1997.03.30 file created as sian.c
            1999.12.28 redesign completed
            2002.03.17 hypertree functions made a separate module
            2013.09.04 adapted to definition of NODEID
            2013.09.11 redesign completed (pointer traversals reduced)
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "arrays.h"
#include "hyper.h"
#ifdef HYP_MAIN
#include <time.h>
#include "random.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define HT_FULL       1         /* hypertree is fully connected */
#define HT_CONSEQ     2         /* construction sequence is valid */

/*----------------------------------------------------------------------
  Auxiliary Functions
----------------------------------------------------------------------*/

static void nodesort (NODEID *nodes, NODEID cnt)
{                               /* --- sort node identifiers */
  NODEID i, t;                  /* loop variable and exchange buffer */
  NODEID *p;                    /* to traverse the node identifiers */

  assert(nodes && (cnt >= 0));  /* check the function arguments */
  if (cnt <= 1) return;         /* sort only more than one identifier */
  for (p = nodes, i = cnt; --i > 0; )  /* find the smallest node */
    if (nodes[i] < *p) p = nodes +i;   /* identifier and swap it */
  t = *nodes; *nodes = *p; *p = t;     /* to front as a sentinel */
  for (i = 0; ++i < cnt; ) {    /* insertion sort with sentinel */
    p = nodes +i; t = *p;       /* note the identifier to insert */
    while (*--p > t) p[1] = *p; /* shift right greater identifiers */
    p[1] = t;                   /* store the identifier to insert */
  }                             /* in the place thus found */
}  /* nodesort() */

/*----------------------------------------------------------------------
Since the number of nodes in a hyperedge will usually be (very) small,
insertion sort is somewhat more efficient than quicksort.
----------------------------------------------------------------------*/

/*----------------------------------------------------------------------
  Hyperedge Functions
----------------------------------------------------------------------*/

static int he_cmp (const void *p1, const void *p2, void *data)
{                               /* --- compare two hyperedges */
  NODEID r1, r2;                /* representative nodes */

  r1 = ((HCOMP*)data)[((const HEDGE*)p1)->nodes[0]].id;
  r2 = ((HCOMP*)data)[((const HEDGE*)p2)->nodes[0]].id;
  if (r1 < r2) return -1;       /* return the sign of the difference */
  return (r1 > r2) ? 1 : 0;     /* of the identifiers of the */
}  /* he_cmp() */               /* representative nodes */

/*--------------------------------------------------------------------*/

static HEDGE* he_rand (HTREE *ht, NODEID size, RANDFN randfn)
{                               /* --- generate a random hyperedge */
  NODEID i, k, n, r, x, z;      /* loop variables, buffers */
  HEDGE  *e;                    /* created hyperedge */

  assert(ht && (size >= 2));    /* check the function arguments */
  if (ht->flags & HT_FULL)      /* if the hypertree is full, */
    return NULL;                /* abort the function */

  /* --- collect isolated nodes --- */
  for (n = 0, i = ht->nodecnt; --i >= 0; ) {
    r = ht->comps[i].id;        /* traverse all nodes */
    if ((r == i) && (ht->comps[i].cnt <= 1))
      ht->nodes[n++] = i;       /* if the node is in no hyperedge, */
    else {                      /* note its identifier, otherwise */
      while ((k = ht->comps[r].id) != r) r = k;
      ht->comps[i].id = r;      /* find the representative node */
    }                           /* and note its identifier */
  }                             /* (flatten the references) */

  /* --- choose one hyperedge per connected component --- */
  memcpy(ht->buf, ht->hedges, (size_t)ht->hecnt *sizeof(HEDGE*));
  ptr_qsort(ht->buf, (size_t)ht->hecnt, +1, he_cmp, ht->comps);
  for (k = 0; k < ht->hecnt; k = x) { /* traverse hyperedge groups */
    r = ht->comps[ht->buf[k]->nodes[0]].id;
    for (x = k; ++x < ht->hecnt; ) /* find hyperedges with same rep. */
      if (ht->comps[ht->buf[x]->nodes[0]].id != r) break;
    i = k +(NODEID)(randfn() *(double)(x-k));
    if (i >= x) i = x-1;        /* compute a random vector index */
    if (i <  k) i = k;          /* (select a random hyperedge) */
    e = ht->buf[i];             /* get the selected hyperedge */
    z = e->size;                /* and copy its nodes */
    memcpy(ht->nodes +n, e->nodes, (size_t)z *sizeof(NODEID));
    i = (NODEID)(randfn() *(double)z);
    if (i > --z) i = z;         /* compute a random vector index */
    if (i <   0) i = 0;         /* (select a random node) */
    ht->nodes[n+i] = ht->nodes[n+z];
    n += z;                     /* delete the selected node */
  }                             /* (one node must be excluded) */

  /* --- create a new hyperedge --- */
  if (size > n) size = n;       /* adapt the hyperedge size */
  e = (HEDGE*)malloc(sizeof(HEDGE) +(size_t)(size-1) *sizeof(NODEID));
  if (!e) return NULL;          /* allocate a memory block and */
  e->size = size;               /* initialize the structure */

  /* --- select nodes of new hyperedge --- */
  for (k = size; --k > 0; ) {   /* node selection loop */
    i = (NODEID)(randfn() *(double)n);
    if (i > --n) i = n;         /* compute a random vector index */
    if (i <   0) i = 0;         /* (select a random node) */
    e->nodes [k] = ht->nodes[i];/* get identifier at the index */
    ht->nodes[i] = ht->nodes[n];/* and replace it with last id. */
  }                             /* (compact identifier list) */
  r = ht->comps[e->nodes[size-1]].id;  /* get a representative node */
  if (ht->comps[r].cnt >= size){/* if single con. comp. is possible */
    for (k = size-1; --k > 0; ) /* traverse the selected nodes */
      if (ht->comps[e->nodes[k]].id != r)
        break;                  /* check for same representative node */
    if (k <= 0) {               /* if there is only one component */
      for (k = n, i = n = 0; i < k; i++)
        if (ht->comps[ht->nodes[i]].id != r)
          ht->nodes[n++] = ht->nodes[i];
      assert(n > 0);            /* (enforce two connected components) */
    }                           /* compute new number of selectable */
  }                             /* nodes and check this number */
  i = (NODEID)(randfn() *(double)n);
  if (i > --n) i = n;           /* compute a random vector index */
  if (i <   0) i = 0;           /* (select a random node) */
  e->nodes[0] = ht->nodes[i];   /* get the identifier at the index */
  nodesort(e->nodes, e->size);  /* sort the selected nodes */
  return e;                     /* return the created hyperedge */
}  /* he_rand() */

/*--------------------------------------------------------------------*/

static HEDGE* he_clone (HEDGE *hedge)
{                               /* --- clone hyperedge */
  HEDGE *clone;                 /* created clone */

  assert(hedge);                /* check the function argument */
  clone = (HEDGE*)malloc(sizeof(HEDGE)
                       +(size_t)(hedge->size-1) *sizeof(NODEID));
  if (!clone) return NULL;      /* allocate a memory block and */
  clone->size = hedge->size;    /* copy size and node identifiers */
  memcpy(clone->nodes,hedge->nodes, (size_t)hedge->size*sizeof(NODEID));
  return clone;                 /* return the created clone */
}  /* he_clone() */

/*--------------------------------------------------------------------*/

static int he_add (HTREE *ht, HEDGE *hedge)
{                               /* --- add a hyperedge to a hypertree */
  NODEID i, k, n, r, x;         /* loop variables, buffers */

  assert(ht && hedge);          /* check the function arguments */
  if (ht->flags & HT_FULL)      /* if the hypertree is already fully */
    return -1;                  /* connected, abort the function */

  /* --- determine representative nodes --- */
  for (i = 0; i < hedge->size; i++) {
    r = k = hedge->nodes[i];    /* traverse the node identifiers */
    while ((x = ht->comps[r].id) != r) r = x;
    ht->nodes[i] = ht->comps[k].id = r;
  }                             /* find and sort representative nodes */
  nodesort(ht->nodes, hedge->size);
  for (i = n = 0; ++i < hedge->size; )
    if (ht->nodes[i] != ht->nodes[n]) /* remove duplicate identifiers */
      ht->nodes[++n] = ht->nodes[i];  /* thus find # of con. comps. */
  if (++n < 2) return -2;       /* check for more than one component */

  /* --- add the hyperedge to the hypertree --- */
  ht->hedges[ht->hecnt++] = hedge;
  ht->flags &= ~HT_CONSEQ;      /* clear the construction seq. flag */

  /* --- update connected components --- */
  for (r = ht->nodes[k = 0], i = n; --i >= 0; ) {
    x = ht->nodes[i];           /* traverse representative nodes */
    if (ht->comps[x].cnt > ht->comps[r].cnt) r = x;
    k += ht->comps[x].cnt;      /* find best new representative */
  }                             /* and sum represented nodes */
  ht->comps[r].cnt = k;         /* set number of represented nodes */
  for (i = 0; i < n; i++)       /* and id of new representative node */
    ht->comps[ht->nodes[i]].id = r;
  if (k >= ht->nodecnt) {       /* if all nodes are in one component, */
    ht->flags |= HT_FULL; return 1; }          /* set the 'full' flag */
  return 0;                     /* return the state of the hypertree */
}  /* he_add() */

/*--------------------------------------------------------------------*/

static int he_subset (HEDGE *he1, HEDGE *he2, NODEID *marks)
{                               /* --- check for node set subsumption */
  NODEID n1, n2, d;             /* to traverse the node identifiers */
  int    r = 3;                 /* return code */

  assert(he1 && he2);           /* check the function arguments */
  n1 = he1->size -1;            /* get indices of last nodes */
  n2 = he2->size -1;            /* of each of the two hyperedges */
  if ((n1 >= 0) && (n2 >= 0)) { /* if neither hyperedge is empty */
    while (r != 0) {            /* while subsumption is possible */
      if (marks) {              /* if a marker vector is given */
        while (marks[he1->nodes[n1]] <= 0) { if (--n1 < 0) goto term; }
        while (marks[he2->nodes[n2]] <= 0) { if (--n2 < 0) goto term; }
      }                         /* skip unmarked nodes */
      d = he1->nodes[n1] -he2->nodes[n2];
      if      (d > 0) { r &= ~1; if (--n1 < 0) goto term; }
      else if (d < 0) { r &= ~2; if (--n2 < 0) goto term; }
      else { --n2; if ((--n1 < 0) || (n2 < 0)) goto term; }
    }                           /* if in either hyperedge there is a */
  }                             /* node not contained in the other */
  term:                         /* clear the corresp. subset flag */
  if (r == 0) return 0;         /* if subsumption is imposs., abort */
  if (!marks) {                 /* if no marker vector is given, */
    if (n1 >= 0) r &= ~1;       /* simply check whether */
    if (n2 >= 0) r &= ~2; }     /* there are remaining nodes */
  else {                        /* if a marker vector is given */
    for ( ; n1 >= 0; n1--)      /* check markers of the rem. nodes */
      if (marks[he1->nodes[n1]] > 0) { r &= ~1; break; }
    for ( ; n2 >= 0; n2--)
      if (marks[he2->nodes[n2]] > 0) { r &= ~2; break; }
  }
  return r;                     /* return subsumption flags */
}  /* he_subset() */

/*----------------------------------------------------------------------
The above function presupposes that the node identifiers are sorted
in ascending order (this is ensured with the function nodesort).
In the return code bit 0 is set if he1 \subseteq he2 and bit 1 is
set if he2 \subseteq he1. If he1 = he2, then both bits are set.
----------------------------------------------------------------------*/

static int he_check (HTREE *ht, HEDGE *hedge)
{                               /* --- check hyperedge */
  int    t;                     /* test result */
  NODEID i, k, n, r, s;         /* loop variables, buffers */
  HEDGE  *e;                    /* to traverse hyperedges */

  assert(hedge && ht);          /* check the function arguments */
  if (ht->flags & HT_FULL)      /* if the hypertree is already fully */
    return -1;                  /* connected, abort the function */
  for (k = hedge->size; --k >= 0; ) {
    r = s = hedge->nodes[k];    /* traverse the hyperedge nodes */
    while ((i = ht->comps[r].id) != r) r = i;
    ht->comps[s].id = r;        /* flatten the references */
    if (ht->comps[r].cnt > 1)   /* if node is still unmarked and */
      ht->comps[r].cnt = -ht->comps[r].cnt;      /* not isolated */
  }                             /* init. representative node markers */
  for (i = ht->hecnt; --i >= 0; ) {
    e = ht->hedges[i];          /* traverse existing hyperedges */
    n = e->size;                /* get the hyperedge size */
    r = s = e->nodes[--n];      /* and its last node */
    while ((k = ht->comps[r].id) != r) r = k;
    ht->comps[s].id = r;        /* flatten the reference */
    if (ht->comps[r].cnt > 0)   /* if the connected component already */
      continue;                 /* passed the test, skip hyperedge */
    for (k = hedge->size; --k >= 0; ) {
      s = hedge->nodes[k];      /* traverse the nodes */
      if (ht->comps[s].id != r) /* in the hyperedge to check */
        continue;               /* from the same connected component */
      while ((e->nodes[n] > s) && (n > 0))
        n--;                    /* search each node of the hyperedge */
      if (e->nodes[n] != s) break;
    }                           /* if not found, abort the loop */
    if (k < 0)                  /* if all nodes have been found */
      ht->comps[r].cnt = -ht->comps[r].cnt;
  }                             /* unmark the representative node */
  for (t = 0, k = 0; k < hedge->size; k++) {
    r = ht->comps[hedge->nodes[k]].id;
    n = ht->comps[r].cnt;       /* traverse nodes of the hyperedge */
    if (n < 0) {                /* if a component did not pass the */
      ht->comps[r].cnt = -n; t = -1; }   /* test, clear marker and */
  }                             /* set the test result to 'failure' */
  return t;                     /* return the result of the test */
}  /* he_check() */

/*----------------------------------------------------------------------
  Hypertree Functions
----------------------------------------------------------------------*/

HTREE* ht_create (NODEID nodecnt)
{                               /* --- create a hypertree */
  HTREE  *ht;                   /* created hypertree */

  assert(nodecnt > 0);          /* check the function argument */
  ht = (HTREE*)malloc(sizeof(HTREE)
                    +(size_t)(2*nodecnt-1) *sizeof(HEDGE*));
  if (!ht) return NULL;         /* allocate the base structure */
  ht->hecnt   = 0;              /* and initialize it */
  ht->buf     = ht->hedges +nodecnt;
  ht->comps   = (HCOMP*) malloc((size_t)nodecnt *sizeof(HCOMP));
  ht->nodes   = (NODEID*)malloc((size_t)nodecnt *sizeof(NODEID));
  if (!ht->comps || !ht->nodes) { ht_delete(ht); return NULL; }
  ht->nodecnt = nodecnt;        /* allocate and init. the vectors */
  ht->flags   = HT_CONSEQ | ((nodecnt < 2) ? HT_FULL : 0);
  while (--nodecnt >= 0) {      /* initialize connected components */
    ht->comps[nodecnt].id = nodecnt; ht->comps[nodecnt].cnt = 1; }
  return ht;                    /* return the created hypertree */
}  /* ht_create() */

/*--------------------------------------------------------------------*/

void ht_delete (HTREE *ht)
{                               /* --- delete a hypertree */
  NODEID i;                     /* loop variable */

  assert(ht);                   /* check the function argument */
  for (i = 0; i < ht->hecnt; i++)
    free(ht->hedges[i]);        /* delete all hyperedges */
  if (ht->nodes) free(ht->nodes);
  if (ht->comps) free(ht->comps);
  free(ht);                     /* delete vectors and base structure */
}  /* ht_delete() */

/*--------------------------------------------------------------------*/

HTREE* ht_clone (HTREE *ht)
{                               /* --- clone hypertree */
  NODEID i;                     /* loop variable */
  HTREE  *clone;                /* clone of hypertree */

  assert(ht);                   /* check the function argument */
  clone = ht_create(ht->nodecnt);          /* create a hypertree */
  if (!clone) return NULL;      /* with the same number of nodes */
  clone->flags = ht->flags;     /* and copy connected components */
  memcpy(clone->comps, ht->comps, (size_t)ht->hecnt *sizeof(NODEID));
  for (i = 0; i < ht->hecnt; i++) {
    clone->hedges[i] = he_clone(ht->hedges[i]);
    if (!clone->hedges[i]) { ht_delete(clone); return NULL; }
    clone->hecnt++;             /* copy and count the hyperedges */
  }
  return clone;                 /* return the created clone */
}  /* ht_clone() */

/*--------------------------------------------------------------------*/

int ht_headd (HTREE *ht, NODEID *nodes, NODEID size)
{                               /* --- add a hyperedge to a hypertree */
  HEDGE *e;                     /* created hyperedge */

  assert(ht && nodes && (size > 1)); /* check the function arguments */
  if (ht->flags & HT_FULL)      /* if the hypertree is already fully */
    return -2;                  /* connected, abort the function */
  e = (HEDGE*)malloc(sizeof(HEDGE)+(size_t)(size-1)*sizeof(NODEID));
  if (!e) return -1;            /* create a hyperedge and */
  e->size = size;               /* note the number of nodes */
  memcpy(e->nodes, nodes, (size_t)size *sizeof(NODEID));
  return he_add(ht, e);         /* copy the node identifiers and */
}  /* ht_headd() */             /* add the hyperedge to the hypertree */

/*--------------------------------------------------------------------*/

int ht_reduce (HTREE *ht)
{                               /* --- Graham reduction */
  NODEID s, d, k, r;            /* loop variables */
  NODEID n;                     /* number of remaining hyperedges */
  int    subset;                /* subset flags (result of he_subset) */
  HEDGE  *e;                    /* to traverse the hyperedges */

  assert(ht);                   /* check the function argument */
  if (ht->flags & HT_CONSEQ)    /* check for a valid */
    return 0;                   /* construction sequence */

  /* --- set up node markers --- */
  for (k = 0; k < ht->nodecnt; k++)
    ht->nodes[k] = -1;          /* initialize hyperedge counters */
  for (s = 0; s < ht->hecnt; s++) {
    e = ht->hedges[s];          /* traverse the hyperedges */
    for (k = 0; k < e->size; k++)
      ht->nodes[e->nodes[k]]++; /* traverse the nodes and count */
  }                             /* the hyperedge for each node */
  /* It is ht->nodes[i] = n if there are n+1 hyperedges containing  */
  /* the node with the identifier i. Hence ht->nodes[i] indicates   */
  /* whether the node is to be checked in the hyperedge subsumption */
  /* test (ht->nodes[i] > 0) or not (ht->nodes[i] <= 0).            */

  /* --- Graham reduction --- */
  for (r = 0, n = ht->hecnt; n > 1; n = d) {
    e = NULL;                   /* clear edge as a flag */
    for (s = n; --s > 0; ) {    /* traverse remaining hyperedges */
      if (!ht->hedges[s]) continue;
      for (d = s; --d >= 0;) {  /* traverse preceding hyperedges */
        if (!ht->hedges[d]) continue;
        subset = he_subset(ht->hedges[s], ht->hedges[d], ht->nodes);
        if (!subset) continue;  /* if no subset relation, skip pair */
        e = ht->hedges[k = (subset & 1) ? s : d];
        ht->hedges[k] = NULL;   /* remove hyperedge that is a subset */
        ht->buf[r++]  = e;      /* and store it in the buffer */
        for (k = 0; k < e->size; k++) /* remove the deleted edge */
          ht->nodes[e->nodes[k]]--;   /* from the counters */
        if (subset & 1) break;  /* if the source has been removed, */
      }                         /* abort the inner loop */
    }                           /* if no hyperedges were removed, */
    if (!e) break;              /* the reduction failed, so abort */
    for (s = d = 0; s < n; s++) /* expunge the deleted entries */
      if (ht->hedges[s]) ht->hedges[d++] = ht->hedges[s];
  }                             /* (collect remaining hyperedges) */

  /* --- copy back removed hyperedges --- */
  for (d = n; --r >= 0; d++)    /* copy the removed hyperedges */
    ht->hedges[d] = ht->buf[r]; /* in reverse order */
  /* If the reduction succeeded, the hyperedges are ordered */
  /* in such a way that they form a construction sequence.  */
  if (n <= 1) ht->flags |= HT_CONSEQ;
  return (n > 1) ? -1 : 0;      /* set the construction sequence flag */
}  /* ht_reduce() */            /* and return the reduction result */

/*--------------------------------------------------------------------*/

int ht_fill (HTREE *ht, NODEID maxsize, RANDFN randfn)
{                               /* --- fill with random hyperedges */
  NODEID size;                  /* size of current hyperedge */
  HEDGE  *e;                    /* current hyperedge */
  int    r;                     /* result of adding a hyperedge */

  assert(ht && (maxsize > 1));  /* check the function arguments */
  if (ht->flags & HT_FULL)      /* if the hypertree is already fully */
    return 0;                   /* connected, abort the function */
  do {                          /* hyperedge adding loop */
    size = (NODEID)(randfn() *(double)(maxsize-1)) +2;
    if (size > maxsize) size = maxsize;
    if (size < 2)       size = 2;  /* choose a size at random */
    e = he_rand(ht, size, randfn);
    if (!e) return -1;          /* generate a random hyperedge */
    r = he_add(ht, e);          /* and add it to the hypertree */
    if (r < 0) { free(e); return -1; }
  } while (r <= 0);             /* while not fully connected */
  ht->flags |= HT_FULL;         /* set the fully connected flag */
  return 0;                     /* return 'ok' */
}  /* ht_fill() */

/*--------------------------------------------------------------------*/

HTREE* ht_modify (HTREE *ht, NODEID maxsize, double keep, RANDFN randfn)
{                               /* --- create modified hypertree */
  NODEID i, k, n;               /* loop variables, buffers */
  HTREE  *mod;                  /* created modified hypertree */
  HEDGE  *e;                    /* to traverse/clone the hyperedges */

  assert(ht && (maxsize >= 0)); /* check the function arguments */
  mod = ht_create(ht->nodecnt); /* create a hypertree */
  if (!mod) return NULL;        /* with the same number of nodes */
  memcpy(ht->buf, ht->hedges, (size_t)(n = ht->hecnt) *sizeof(HEDGE*));
  k = (NODEID)(keep *(double)n);/* compute the number */
  if (k >= n) k = n-1;          /* of hyperedges to keep */
  if (k <  0) k = 0;            /* (it must be 0 <= k < n) */
  while ((n > 0) && (k > 0)) {  /* select random hyperedges */
    i = (NODEID)(randfn() *(double)n);
    if (i > --n) i = n;         /* compute a random index and */
    if (i <   0) i = 0;         /* check and adapt its range */
    e          = ht->buf[i];    /* select a random hyperedge and */
    ht->buf[i] = ht->buf[n];    /* fill the gap with the last h.e. */
    if (he_check(mod, e) != 0)  /* check whether the hyperedge */
      continue;                 /* can be added to the tree */
    e = he_clone(e);            /* clone the selected hyperedge */
    if (!e) {  ht_delete(mod); return NULL; }
    if (he_add(mod, e) < 0) {   /* add the cloned hyperedge */
      free(e); ht_delete(mod); return NULL; }
    k--;                        /* decrement the hyperedge counter */
  }
  if (ht_fill(mod, maxsize, randfn) < 0) {
    ht_delete(mod); return NULL; }  /* fill with random hyperedges */
  return mod;                   /* return the created hypertree */
}  /* ht_modify() */

/*--------------------------------------------------------------------*/
#if !defined NDEBUG || defined HYP_MAIN

void ht_show (HTREE *ht)
{                               /* --- show a hypertree */
  NODEID i, k;                  /* loop variables */
  HEDGE  *e;                    /* to traverse the hyperedges */

  assert(ht);                   /* check the function argument */
  for (i = 0; i < ht->hecnt; i++) {
    e = ht->hedges[i];          /* traverse the hyperedges */
    printf("(");                /* start the hyperedge */
    for (k = 0; k < e->size; k++) {
      if (k > 0) printf(",");   /* print a separating comma */
      printf("%"NODEID_FMT, e->nodes[k]);
    }                           /* print node identifiers */
    printf(")\n");              /* terminate the hyperedge */
  }
}  /* ht_show() */

#endif
/*--------------------------------------------------------------------*/
#ifdef HYP_MAIN

int main (void)
{                               /* --- main function for testing */
  HTREE  *ht = NULL;            /* test hypertree */
  NODEID ids[8];                /* node identifiers of hyperedge */

  rseed((unsigned)time(NULL));
  ht = ht_create(8);
  printf("created hypertree with %d nodes\n", ht_nodecnt(ht));
  ht_fill(ht, 8, drand);
  ht_show(ht);
  printf("reduce: %d\n", ht_reduce(ht));
  ht_show(ht);
  ht_delete(ht);

  ht = ht_create(6);
  printf("created hypertree with %d nodes\n", ht_nodecnt(ht));
  ids[0] = 0; ids[1] = 1; ids[2] = 2;
  ht_headd(ht, ids, 3);
  ids[0] = 0; ids[1] = 2; ids[2] = 3;
  ht_headd(ht, ids, 3);
  ids[0] = 0; ids[1] = 3; ids[2] = 4;
  ht_headd(ht, ids, 3);
  ids[0] = 0; ids[1] = 4; ids[2] = 1, ids[3] = 5;
  ht_headd(ht, ids, 4);
  ht_show(ht);
  printf("reduce: %d\n", ht_reduce(ht));

  return 0;
}  /* main() */

#endif
