/*----------------------------------------------------------------------
  File    : ptree3.c
  Contents: probability/possibility tree management
            (conditional) dependence measurement functions
  Author  : Christian Borgelt
  History : 1995.10.22 file created as ctree.c
            1997.03.25 function pt_depend() added
            2002.01.02 switched to sorting function from arrays
            2002.01.08 symmetric Gini index added
            2002.01.12 normalized chi^2 measure added
            2002.01.22 adapted to redesigned node structure
            2002.02.04 quadratic information measures added
            2003.01.17 normalization of possibilistic measures added
            2006.03.05 evaluation result stored in field 'eval'
            2007.01.29 bug in function phi2() fixed (factor n)
            2013.09.04 adapted to definitions ATTID, VALID, TPLID etc.
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#include "arrays.h"
#include "ptree.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define LN_2   0.693147180559945   /* ln(2) */

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- measure information --- */
  char *name;                   /* name of the measure */
  int  flags;                   /* flags, e.g. PT_SYM */
} MINFO;                        /* (measure information) */

typedef double MARGFN (PTNODE *node, PTLVL *lvl);
typedef double DEPFN  (PTNODE *node, PTLVL *lvl,
                       int measure, double *params);

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
static const MINFO mi_prob[PT_UNKNOWN+1] = {
  /* PT_NONE     0 */ { "no measure",                               1 },
  /* PT_INFGAIN  1 */ { "information gain",                    PT_SYM },
  /* PT_INFGBAL  2 */ { "balanced information gain",                1 },
  /* PT_INFGR    3 */ { "information gain ratio",                   1 },
  /* PT_INFSGR1  4 */ { "symmetric information gain ratio 1",  PT_SYM },
  /* PT_INFSGR2  5 */ { "symmetric information gain ratio 2",  PT_SYM },
  /* PT_QIGAIN   6 */ { "quadratic information gain",          PT_SYM },
  /* PT_QIGBAL   7 */ { "balanced quadratic information gain",      1 },
  /* PT_QIGR     8 */ { "quadratic information gain ratio",         1 },
  /* PT_QISGR1   9 */ { "symmetric quad. info. gain ratio 1",  PT_SYM },
  /* PT_QISGR2  10 */ { "symmetric quad. info. gain ratio 2",  PT_SYM },
  /* PT_GINI    11 */ { "Gini index",                               1 },
  /* PT_GINISYM 12 */ { "symmetric Gini index",                PT_SYM },
  /* PT_GINIMOD 13 */ { "modified Gini index",                      1 },
  /* PT_RELIEF  14 */ { "relief measure",                           1 },
  /* PT_WDIFF   15 */ { "sum of weighted differences",         PT_SYM },
  /* PT_CHI2    16 */ { "chi^2 measure",                       PT_SYM },
  /* PT_CHI2NRM 17 */ { "normalized chi^2 measure",            PT_SYM },
  /* PT_WEVID   18 */ { "weight of evidence",                       1 },
  /* PT_RELEV   19 */ { "relevance",                                1 },
  /* PT_BDM     20 */ { "Bayesian-Dirichlet / K2 metric",           1 },
  /* PT_BDMOD   21 */ { "modified Bayesian-Dirichlet / K2 metric",  1 },
  /* PT_RDLREL  22 */ { "reduction of description length "
                          "(rel. freq.)", 1 },
  /* PT_RDLABS  23 */ { "reduction of description length "
                          "(abs. freq.)", 1 },
  /* PT_STOCO   24 */ { "stochastic complexity",                    1 },
  /* PT_UNKNOWN 25 */ { "<unknown>",                                0 }
};                              /* measures for possibility trees */

static const MINFO mi_poss[PT_UNKNOWN+1] = {
  /* PT_NONE     0 */ { "no measure",                               1 },
  /* PT_SPCGAIN  1 */ { "specificity gain",                    PT_SYM },
  /* PT_SPCGBAL  2 */ { "balanced specificity gain",                1 },
  /* PT_SPCGR    3 */ { "specificity gain ratio",                   1 },
  /* PT_SPCSGR1  4 */ { "symmetric specificity gain ratio 1",  PT_SYM },
  /* PT_SPCSGR2  5 */ { "symmetric specificity gain ratio 2",  PT_SYM },
  /*             6 */ { "<unknown>",                                0 },
  /*             7 */ { "<unknown>",                                0 },
  /*             8 */ { "<unknown>",                                0 },
  /*             9 */ { "<unknown>",                                0 },
  /*            10 */ { "<unknown>",                                0 },
  /*            11 */ { "<unknown>",                                0 },
  /*            12 */ { "<unknown>",                                0 },
  /*            13 */ { "<unknown>",                                0 },
  /*            14 */ { "<unknown>",                                0 },
  /* PT_WDIFF   15 */ { "sum of weighted differences",         PT_SYM },
  /* PT_CHI2    16 */ { "possibilistic chi^2 measure",         PT_SYM },
  /*            17 */ { "<unknown>",                                0 },
  /*            18 */ { "<unknown>",                                0 },
  /*            19 */ { "<unknown>",                                0 },
  /* PT_MUTSPC  20 */ { "mutual specificity",                  PT_SYM },
  /*            21 */ { "<unknown>",                                0 },
  /*            22 */ { "<unknown>",                                0 },
  /*            23 */ { "<unknown>",                                0 },
  /*            24 */ { "<unknown>",                                0 },
  /* PT_UNKNOWN 25 */ { "<unknown>",                                0 }
};                              /* measures for probability trees */

/*----------------------------------------------------------------------
  Functions to Compute Marginals
----------------------------------------------------------------------*/

static double marg_sum (PTNODE *node, PTLVL *lvl)
{                               /* --- marginalize by summing */
  VALID  i, j;                  /* loop variables */
  PTNODE **p;                   /* to traverse the child vector */
  WEIGHT *bp, *bc, *c;          /* to traverse the buffers/counters */
  WEIGHT sum = 0;               /* sum of frequencies */

  assert(node && lvl            /* check the function arguments */
  &&    (lvl[1].index < 0));    /* (next level must be leaves) */
  for (bc = lvl[1].buf +(i = lvl[1].cnt); --i >= 0; )
    *--bc = 0;                  /* clear child marginals */
  p = node->children +lvl->cnt; /* traverse the child vector */
  for (bp = lvl[0].buf +(j = lvl[0].cnt); --j >= 0; ) {
    *--bp = 0;                  /* clear parent marginals */
    if (!*--p) continue;        /* if there is no child, skip value */
    c = ((PTLEAF*)*p)->cnts +lvl[1].cnt;
    for (bc = lvl[1].buf +(i = lvl[1].cnt); --i >= 0; ) {
      *  bp += *--c;            /* aggregate the values */
      *--bc += *  c;            /* of the joint distribution */
    }                           /* in the parent/child buffers */
    sum += *bp;                 /* compute the global total */
  }
  return lvl[0].total = sum;    /* store the total frequency */
}  /* marg_sum() */

/*--------------------------------------------------------------------*/

static double marg_max (PTNODE *node, PTLVL *lvl)
{                               /* --- marginalize by maximizing */
  VALID  i, j;                  /* loop variables */
  PTNODE **p;                   /* to traverse the child vector */
  WEIGHT *bp, *bc, *c;          /* to traverse the buffers/counters */
  WEIGHT max = 0;               /* maximum of frequencies */

  assert(node && lvl            /* check the function arguments */
  &&    (lvl[1].index < 0));    /* (next level must be leaves) */
  for (bc = lvl[1].buf +(i = lvl[1].cnt); --i >= 0; )
    *--bc = 0;                  /* clear child marginals */
  p = node->children +lvl->cnt; /* traverse the child vector */
  for (bp = lvl[0].buf +(j = lvl[0].cnt); --j >= 0; ) {
    *--bp = 0;                  /* clear parent marginals */
    if (!*--p) continue;        /* if there is no child, skip value */
    c = ((PTLEAF*)*p)->cnts +lvl[1].cnt;
    for (bc = lvl[1].buf +(i = lvl[1].cnt); --i >= 0; ) {
      if (*--c > *bp) *bp = *c; /* aggregate the values */
      if (*c > *--bc) *bc = *c; /* of the joint distribution */
    }                           /* in the parent/child buffers */
    if (*bp > max) max = *bp;   /* compute the global maximum */
  }                             /* and store it */
  return lvl[0].total = (double)max;
}  /* marg_max() */

/*----------------------------------------------------------------------
  Probabilistic Conditional Dependence Measures
----------------------------------------------------------------------*/

static double info (PTNODE *node, PTLVL *lvl,
                    int measure, double *params)
{                               /* --- Shannon information measures */
  VALID  i, j;                  /* loop variables */
  PTNODE **p;                   /* to traverse the child vector */
  WEIGHT *c;                    /* to traverse the buffers/counters */
  double s_i, s_j, s_ij;        /* sums for entropy computation */
  double info, t;               /* information gain (ratio), buffer */

  assert(node && lvl);          /* check the function arguments */
  s_i = s_j = s_ij = 0;         /* and initialize the sums */
  for (c = lvl[1].buf +(i = lvl[1].cnt); --i >= 0; )
    if (*--c > 0) s_i += *c *log(*c);
  for (c = lvl[0].buf +(j = lvl[0].cnt); --j >= 0; )
    if (*--c > 0) s_j += *c *log(*c);
  for (p = node->children +(j = lvl[0].cnt); --j >= 0; ) {
    if (!*--p) continue;        /* traverse the existing children */
    t = 0;                      /* process the joint distribution */
    for (c = ((PTLEAF*)*p)->cnts +(i = lvl[1].cnt); --i >= 0; )
      if (*--c > 0) t += *c *log(*c);
    s_ij += t;                  /* process leaves individually and */
  }                             /* sum the results (higher accuracy) */
  t = lvl[0].total; t *= log(t);/* compute N *log(N) only once */
  s_i  = t -s_i;                /* N H_i  = -N sum_i  p_i  *log(p_i)  */
  s_j  = t -s_j;                /* N H_j  = -N sum_j  p_j  *log(p_j)  */
  s_ij = t -s_ij;               /* N H_ij = -N sum_ij p_ij *log(p_ij) */
  info = s_i +s_j -s_ij;        /* compute information gain *N *ln(2) */
  switch (measure) {            /* evaluate measure code */
    case PT_INFSGR1: if (s_ij     <= 0) return 0;
                     info /= s_ij;               break;
    case PT_INFSGR2: if (s_i +s_j <= 0) return 0;
                     info /= s_i +s_j;           break;
    default        : info /= LN_2 *lvl[0].total; break;
  }                             /* form requested entropy ratio */
  return (info > 0) ? info : 0; /* return the information measure */
}  /* info() */

/*--------------------------------------------------------------------*/

static double quad (PTNODE *node, PTLVL *lvl,
                    int measure, double *params)
{                               /* --- quadratic information measures */
  VALID  i, j;                  /* loop variables */
  PTNODE **p;                   /* to traverse the child vector */
  WEIGHT *c;                    /* to traverse the buffers/counters */
  double s_i, s_j, s_ij, s;     /* sums of squared frequencies */
  double quad, t;               /* quadratic information gain, buffer */

  assert(node && lvl);          /* check the function arguments */
  s_i = s_j = s_ij = 0;         /* initialize sums */
  for (c = lvl[1].buf +(i = lvl[1].cnt); --i >= 0; ) {
    t = *--c; s_i += t*t; }     /* compute sum_i N_i^2 (child dist.) */
  for (c = lvl[0].buf +(j = lvl[0].cnt); --j >= 0; ) {
    t = *--c; s_j += t*t; }     /* compute sum_j N_j^2 (parent dist.) */
  for (p = node->children +(j = lvl[0].cnt); --j >= 0; ) {
    if (!*--p) continue;        /* traverse the existing children */
    s = 0;                      /* and the counters of each child */
    for (c = ((PTLEAF*)*p)->cnts +(i = lvl[1].cnt); --i >= 0; ) {
      t = *--c; s += t*t; }     /* compute sum_ij N_ij^2 */
    s_ij += s;                  /* (joint distribution) */
  }
  t = lvl[0].total; t *= t;     /* compute N^2 only once */
  s_i  = t -s_i;                /* N^2/2 H^2_i  = N^2 -sum_i  N_i^2  */
  s_j  = t -s_j;                /* N^2/2 H^2_j  = N^2 -sum_j  N_j^2  */
  s_ij = t -s_ij;               /* N^2/2 H^2_ij = N^2 -sum_ij N_ij^2 */
  quad = s_i +s_j -s_ij;        /* compute quad. info. gain *N^2/2 */
  switch (measure) {            /* evaluate measure code */
    case PT_QISGR1: if (s_ij     <= 0) return 0;
                    quad /= s_ij;      break;
    case PT_QISGR2: if (s_i +s_j <= 0) return 0;
                    quad /= s_i +s_j;  break;
    default       : quad /= 0.5 *t;    break;
  }                             /* form requested entropy ratio */
  return (quad > 0) ? quad : 0; /* return the information measure */
}  /* quad() */

/*--------------------------------------------------------------------*/

static double gini (PTNODE *node, PTLVL *lvl,
                    int measure, double *params)
{                               /* --- symmetric Gini index */
  VALID  i, j;                  /* loop variables */
  PTNODE **p;                   /* to traverse the child vector */
  WEIGHT *bp, *bc, *bx, *c;     /* to traverse the buffers/counters */
  double s_i, s_j, s;           /* sums of squared frequencies */
  double w_ij, w_ji;            /* weighted sums of squared freqs. */
  double gini, n, t;            /* symmetric Gini index, buffer */

  assert(node && lvl);          /* check the function arguments */
  s_i = s_j = w_ij = 0;         /* initialize sums */
  for (bc = lvl[1].buf +(i = lvl[1].cnt); --i >= 0; ) {
    t = *--bc; s_i += t*t; }    /* compute sum_i N_i^2 (child dist.) */
  for (bp = lvl[0].buf +(j = lvl[0].cnt); --j >= 0; ) {
    t = *--bp; s_j += t*t; }    /* compute sum_j N_j^2 (parent dist.) */
  for (bx = bc +lvl[1].cnt +(i = lvl[1].cnt); --i >= 0; )
    *--bx = 0;                  /* clear the extended buffer */
  for (p = node->children  +(j = lvl[0].cnt); --j >= 0; ) {
    if (!*--p) continue;        /* traverse the existing children */
    s = 0;                      /* and the counters of each child */
    for (c = ((PTLEAF*)*p)->cnts +(i = lvl[1].cnt); --i >= 0; ) {
      t = *--c; s += t *= t;              /* compute N_ij^2 and */
      bx[i] = (WEIGHT)((double)bx[i] +t); /* sum N_ij^2 over i and */
    }                                     /* (in buffer bx) over j */
    if (bp[j] > 0) w_ij += s /bp[j];
  }                             /* compute weighted sum over j */
  w_ji = 0;                     /* and afterwards also over i */
  for (i = lvl[1].cnt; --i >= 0; )
    if (bc[i] > 0) w_ji += bx[i] /bc[i];
  n    = lvl[0].total;          /* get the total number of cases */
  gini = n *(w_ij +w_ji) -(t = s_i -s_j);
  t    = 2 *n *n -t;            /* compute numerator and denominator */
  return ((gini <= 0) || (t <= 0)) ? 0 : gini/t;
}  /* gini() */                 /* return the symmetric Gini index */

/*--------------------------------------------------------------------*/

static double wdiff (PTNODE *node, PTLVL *lvl,
                     int measure, double *params)
{                               /* --- weighted absolute difference */
  VALID  i, j;                  /* loop variables */
  int    e;                     /* exponent for differences */
  PTNODE **p;                   /* to traverse the child vector */
  WEIGHT bp, *bc, *c;           /* to traverse the buffers/counters */
  double wdiff = 0, n, t;       /* weighted abs. diff., buffers */

  assert(node && lvl && params);/* check the function arguments */
  if      (*params == 1) e = 1; /* note simple exponents */
  else if (*params == 2) e = 2; /* as an integer value */
  else                   e = 0; /* for faster checks */
  n = lvl[0].total;             /* get the total frequency */
  for (p = node->children +(j = lvl[0].cnt); --j >= 0; ) {
    if (!*--p) continue;        /* traverse the existing child nodes */
    bp = lvl[0].buf[j];         /* note the parent frequency */
    bc = lvl[1].buf +(i = lvl[1].cnt);
    c  = ((PTLEAF*)*p)->cnts+i; /* traverse the counters */
    if      (e == 1)            /* if exponent for difference is 1, */
      while (--i >= 0) {        /* sum weighted absolute differences */
        t = *--bc *bp -*--c *n; wdiff += *c *fabs(t); }
    else if (e == 2)            /* if exponent for difference is 2, */
      while (--i >= 0) {        /* sum weighted squared differences */
        t = *--bc *bp -*--c *n; wdiff += *c *t *t; }
    else                        /* if any other exponent, */
      while (--i >= 0) {        /* raise differences to given power */
        t = *--bc *bp -*--c *n; wdiff += *c *pow(fabs(t), *params); }
  }
  t = n*n;                      /* compute the denominator */
  if      (e == 2) t = t*t;     /* of the normalization factor */
  else if (e != 1) t = pow(t, *params);
  return wdiff / (n*t);         /* return the norm. weighted diff. */
}  /* wdiff() */

/*--------------------------------------------------------------------*/

static double chi2 (PTNODE *node, PTLVL *lvl,
                    int measure, double *params)
{                               /* --- chi^2 measure */
  VALID  i, j;                  /* loop variables */
  PTNODE **p;                   /* to traverse the child vector */
  WEIGHT bp, *bc, *c;           /* to traverse the buffers/counters */
  double chi2 = 0, n, x, y;     /* chi^2 measure, buffers */

  assert(node && lvl);          /* check the function arguments */
  n = lvl[0].total;             /* get the total frequency */
  for (p = node->children +(j = lvl[0].cnt); --j >= 0; ) {
    if (!*--p) continue;        /* traverse the existing children */
    bp = lvl[0].buf[j];         /* get the parent frequency */
    bc = lvl[1].buf +lvl[1].cnt;
    for (c = ((PTLEAF*)*p)->cnts +(i = lvl[1].cnt); --i >= 0; ) {
      --c; y = *--bc *bp;       /* traverse the child frequencies */
      if (y <= 0) continue;     /* and check the denominator */
      x = y - *c *n;            /* compute the numerator and */
      chi2 += x*x /y;           /* from it the next term */
    }                           /* of the chi^2 measure */
  }
  if (measure == PT_CHI2NRM) {  /* if normalized version requested */
    x = ((double)lvl[0].cnt -1) *((double)lvl[1].cnt -1);
    if (x > 0) chi2 /= x;       /* normalize the chi12 measure by */
  }                             /* dividing by the degrees of freedom */
  return (chi2 > 0) ? chi2 / (n*n) : 0;
}  /* chi2() */                 /* return the norm. chi^2 measure */

/*----------------------------------------------------------------------
  Possibilistic Conditional Dependence Measures
----------------------------------------------------------------------*/

static double spec (PTNODE *node, PTLVL *lvl,
                    int measure, double *params)
{                               /* --- specificity measures */
  VALID  i, j;                  /* loop variables */
  PTNODE **p;                   /* to traverse the child vector */
  WEIGHT *buf, *b, *c;          /* to traverse the buffers/counters */
  double s_i, s_j, s_ij;        /* nonspecificity of distributions */
  double spec;                  /* specificity gain (ratio) */

  assert(node && lvl);          /* check the function arguments */
  buf = b = (WEIGHT*)malloc((size_t)lvl[0].cnt
                           *(size_t)lvl[1].cnt *sizeof(WEIGHT));
  if (!buf) return NAN;         /* allocate workspace */
  for (p = node->children +(j = lvl[0].cnt); --j >= 0; ) {
    if (!*--p) continue;        /* traverse the existing children */
    for (c = ((PTLEAF*)*p)->cnts +(i = lvl[1].cnt); --i >= 0; )
      if (*--c > 0) *b++ = *c;  /* copy the non-zero counters */
  }                             /* of the joint distribution */
  s_j  = pt_nsp(lvl[0].buf, lvl[0].cnt);
  s_i  = pt_nsp(lvl[1].buf, lvl[1].cnt);
  s_ij = pt_nsp(buf, (VALID)(b -buf));
  free(buf);                    /* compute the different specs. and */
  spec = s_i +s_j -s_ij;        /* from them the specificity gain */
  switch (measure) {            /* evaluate the measure code */
    case PT_SPCSGR1: if (s_ij     <= 0) return 0;
                     spec /= s_ij;               break;
    case PT_SPCSGR2: if (s_i +s_j <= 0) return 0;
                     spec /= s_i +s_j;           break;
    default        : spec /= LN_2 *lvl[1].total; break;
  }                             /* form requested specificity ratio */
  return (spec > 0) ? spec : 0; /* return specificity measure */
}  /* spec() */

/*--------------------------------------------------------------------*/

static double pdiff (PTNODE *node, PTLVL *lvl,
                     int measure, double *params)
{                               /* --- poss. weighted abs. difference */
  VALID  i, j;                  /* loop variables */
  int    e;                     /* exponent for differences */
  PTNODE **p;                   /* to traverse the child vector */
  WEIGHT bp, *bc, *c;           /* to traverse the buffers/counters */
  double pdiff = 0, t;          /* weighted abs. diff., buffers */

  assert(node && lvl);          /* check the function arguments */
  if      (*params == 1) e = 1; /* note simple exponents */
  else if (*params == 2) e = 2; /* as an integer value */
  else                   e = 0; /* for faster checks */
  for (p = node->children +(j = lvl[0].cnt); --j >= 0; ) {
    if (!*--p) continue;        /* traverse the existing children */
    bp = lvl[0].buf[j];         /* get the parent frequency */
    bc = lvl[1].buf +(i = lvl[1].cnt);
    c  = ((PTLEAF*)*p)->cnts+i; /* traverse the counters */
    if      (e == 1)            /* if exponent for difference is 1, */
      while (--i >= 0) {        /* sum weighted absolute differences */
        t = ((*--bc < bp) ? *bc : bp) -*--c; pdiff += *c *fabs(t); }
    else if (e == 2)            /* if exponent for difference is 2, */
      while (--i >= 0) {        /* sum weighted squared differences */
        t = ((*--bc < bp) ? *bc : bp) -*--c; pdiff += *c *t *t; }
    else {                      /* if any other exponent, */
      while (--i >= 0) {        /* raise differences to given power */
        t = ((*--bc < bp) ? *bc : bp) -*--c;
        pdiff += *c *pow(fabs(t), *params);
      }                         /* (a call to the function pow() */
    }                           /* is costly, so it is avoided for */
  }                             /* the simple exponents 1 and 2) */
  return pdiff /lvl[1].total;   /* return the weighted difference */
}  /* pdiff() */

/*--------------------------------------------------------------------*/

static double pchi2 (PTNODE *node, PTLVL *lvl,
                     int measure, double *params)
{                               /* --- possibilistic chi^2 measure */
  VALID  i, j;                  /* loop variables */
  PTNODE **p;                   /* to traverse the child vector */
  WEIGHT bp, *bc, *c;           /* to traverse the buffers/counters */
  double chi2 = 0, x, y;        /* chi^2 measure, buffers */

  assert(node && lvl);          /* check the function arguments */
  for (p = node->children +(i = lvl[0].cnt); --i >= 0; ) {
    if (!*--p) continue;        /* traverse the existing children */
    bp = lvl[0].buf[i];         /* get the parent frequency */
    bc = lvl[1].buf +lvl[1].cnt;
    for (c = ((PTLEAF*)*p)->cnts +(j = lvl[1].cnt); --j >= 0; ) {
      --c; y = (*--bc < bp) ? *bc : bp;
      if (y <= 0) continue;     /* traverse the child frequencies, */
      x = y - *c;               /* compute the denominator and the */
      chi2 += x*x /y;           /* numerator of the next term */
    }                           /* and sum these quotients */
  }                             /* return the poss. chi^2 measure */
  return (chi2 > 0) ? chi2 /lvl[1].total : 0;
}  /* pchi2() */

/*--------------------------------------------------------------------*/

static double mutspc (PTNODE *node, PTLVL *lvl,
                      int measure, double *params)
{                               /* --- mutual specificity */
  VALID  i, j;                  /* loop variables */
  PTNODE **p;                   /* to traverse the child vector */
  WEIGHT bp, *bc, *c;           /* to traverse the buffers */
  double mutspc = 0;            /* mutual specificity */

  assert(node && lvl);          /* check the function arguments */
  for (p = node->children +(j = lvl[0].cnt); --j >= 0; ) {
    if (!*--p) continue;        /* traverse the existing children */
    bp = lvl[0].buf[j];         /* get the parent frequency */
    bc = lvl[1].buf +lvl[1].cnt -1;
    for (c = ((PTLEAF*)*p)->cnts +(i = lvl[1].cnt); --i >= 0; bc--) {
      if (*--c <= 0) continue;  /* traverse the existing children */
      mutspc -= *c *log(*c / ((*bc < bp) ? *bc : bp));
    }                           /* traverse the child frequencies */
  }                             /* and compute the mutual specificity */
  return (mutspc > 0) ? mutspc /lvl[1].total : 0;
}  /* mutspc() */               /* return the mutual specificity */

/*----------------------------------------------------------------------
  Tables of Evaluation Functions
----------------------------------------------------------------------*/
static DEPFN *dep_prob[PT_UNKNOWN+1] = {
  /* PT_NONE     0 */  (DEPFN*)0,
  /* PT_INFGAIN  1 */  info,
  /* PT_INFGBAL  2 */  (DEPFN*)0,
  /* PT_INFGR    3 */  (DEPFN*)0,
  /* PT_INFSGR1  4 */  info,
  /* PT_INFSGR2  5 */  info,
  /* PT_QIGAIN   6 */  quad,
  /* PT_QIGBAL   7 */  (DEPFN*)0,
  /* PT_QIGR     8 */  (DEPFN*)0,
  /* PT_QISGR1   9 */  quad,
  /* PT_QISGR2  10 */  quad,
  /* PT_GINI    11 */  (DEPFN*)0,
  /* PT_GINISYM 12 */  gini,
  /* PT_GINI2   13 */  (DEPFN*)0,
  /* PT_RELIEF  14 */  (DEPFN*)0,
  /* PT_WDIFF   15 */  wdiff,
  /* PT_CHI2    16 */  chi2,
  /* PT_CHI2NRM 17 */  chi2,
  /* PT_WOEVID  18 */  (DEPFN*)0,
  /* PT_RELEV   19 */  (DEPFN*)0,
  /* PT_BDM     20 */  (DEPFN*)0,
  /* PT_BDMOD   21 */  (DEPFN*)0,
  /* PT_RDLEN1  22 */  (DEPFN*)0,
  /* PT_RDLEN2  23 */  (DEPFN*)0,
  /* PT_STOCO   24 */  (DEPFN*)0,
  /* PT_UNKNOWN 25 */  (DEPFN*)0
};

static DEPFN *dep_poss[PT_UNKNOWN+1] = {
  /* PT_NONE     0 */  (DEPFN*)0,
  /* PT_SPCGAIN  1 */  spec,
  /* PT_SPCGBAL  2 */  (DEPFN*)0,
  /* PT_SPCGR    3 */  (DEPFN*)0,
  /* PT_SPCSGR1  4 */  spec,
  /* PT_SPCSGR2  5 */  spec,
  /*             6 */  (DEPFN*)0,
  /*             7 */  (DEPFN*)0,
  /*             8 */  (DEPFN*)0,
  /*             9 */  (DEPFN*)0,
  /*            10 */  (DEPFN*)0,
  /*            11 */  (DEPFN*)0,
  /*            12 */  (DEPFN*)0,
  /*            13 */  (DEPFN*)0,
  /*            14 */  (DEPFN*)0,
  /* PT_WDIFF   15 */  pdiff,
  /* PT_PCHI2   16 */  pchi2,
  /*            17 */  (DEPFN*)0,
  /*            18 */  (DEPFN*)0,
  /*            19 */  (DEPFN*)0,
  /* PT_MUTSPC  20 */  mutspc,
  /*            21 */  (DEPFN*)0,
  /*            22 */  (DEPFN*)0,
  /*            23 */  (DEPFN*)0,
  /*            24 */  (DEPFN*)0,
  /* PT_UNKNOWN 25 */  (DEPFN*)0
};

/*----------------------------------------------------------------------
  Main Functions
----------------------------------------------------------------------*/

double pt_depend (PTREE *pt, int measure, double *params)
{                               /* --- measure cond. independence */
  PTNODE *node;                 /* to traverse the nodes */
  PTLVL  *lvl;                  /* leaf level */
  MARGFN *margfn;               /* marginalization function */
  DEPFN  *depfn;                /* dependence measurement function */
  double n, sum;                /* weight of distribs. and their sum */
  double res;                   /* evaluation result */

  assert(pt);                   /* check the function arguments */
  if ((pt->concnt <= 0)         /* check the tree height */
  ||  (measure < PT_NONE) || (measure >= PT_UNKNOWN))
    return NAN;                 /* check the measure identifier */
  if (pt->type == PT_PROB) {    /* if probability tree */
    margfn = marg_sum; depfn = dep_prob[measure]; }
  else {                        /* if possibility tree */
    margfn = marg_max; depfn = dep_poss[measure]; }
  if (!depfn) return NAN;       /* get marg. and meas. functions */
  lvl = pt->levels +pt->concnt -1;
  res = sum = n = 0;            /* get the level above the leaves */
  if (pt->type == PT_POSS)      /* note the normalization factor */
    lvl[1].total = (pt->tplcnt > 0) ? pt->tplcnt : 1;
  for (node = lvl->list; node; node = node->succ) {
    n = margfn(node, lvl);      /* marginalize the cond. distribution */
    if (n <= 0) continue;       /* skip empty distributions */
    sum += n;                   /* sum the distribution weight */
    res += n * depfn(node, lvl, measure, params);
  }                             /* check for cond. (in)dependence */
  return pt->eval = (sum > 0) ? res/sum : 0;
}  /* pt_depend() */            /* return the evaluation result */

/*--------------------------------------------------------------------*/

const char* pt_mname (int type, int measure)
{                               /* --- get name of evaluation measure */
  if ((measure < PT_NONE) || (measure > PT_UNKNOWN))
    measure = PT_UNKNOWN;       /* check the measure index */
  return ((type == PT_PROB) ? mi_prob : mi_poss)[measure].name;
}  /* pt_mname() */             /* return the measure name */

/*--------------------------------------------------------------------*/

int pt_minfo (int type, int measure)
{                               /* --- get info about eval. measure */
  if ((measure < PT_NONE) || (measure > PT_UNKNOWN))
    measure = PT_UNKNOWN;       /* check the measure index */
  return ((type == PT_PROB) ? mi_prob : mi_poss)[measure].flags;
}  /* pt_minfo() */             /* return the measure information */

/*--------------------------------------------------------------------*/

double pt_nsp (WEIGHT *dist, VALID n)
{                               /* --- compute nonspecificity */
  double nsp  = 0;              /* nonspecificity */
  WEIGHT prec = 0;              /* preceding frequency */
  WEIGHT t;                     /* temporary buffer */

  assert(dist && (n >= 0));     /* check the function arguments */
  wgt_qsort(dist,(size_t)n,+1); /* sort the frequencies and */
  for ( ; n > 1; n--) {         /* then traverse them */
    t = *dist -prec; prec = *dist++;
    if (t > 0) nsp += t *log((double)n);
  }                             /* compute the nonspecificity of */
  return nsp;                   /* the distribution and return it */
}  /* pt_nsp() */

/*----------------------------------------------------------------------
The above auxiliary function is placed here, because it is needed in
ptree3.c, ptree4.c, and ptree5.c to compute the specificity measures.
----------------------------------------------------------------------*/
